#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re

PATTERN_INVALID_CHARACTERS = r"[^a-z0-9\-]"
REGEX_INVALID_CHARACTERS = re.compile(PATTERN_INVALID_CHARACTERS, re.I)

#: Regular expression for matching a valid serial number
PATTERN_VALID_SERIAL = r"^([a-z0-9]|[a-z0-9]+[a-z0-9\-]*[a-z0-9])$"
REGEX_VALID_SERIAL = re.compile(PATTERN_VALID_SERIAL, re.I)


def mangled_serial_number(value, strict=False):
    """
    Mangle input for use as serial number

    Args:
        value (str): input value

    Returns:
        str: mangled

    >>> mangled_serial_number("123")
    '123'
    >>> mangled_serial_number(123)
    '123'
    >>> mangled_serial_number(0)
    '0'
    >>> mangled_serial_number(0.0)
    '00'
    >>> mangled_serial_number(None)
    Traceback (most recent call last):
        ...
    ValueError: None
    >>> mangled_serial_number("")
    Traceback (most recent call last):
        ...
    ValueError: ''
    >>> mangled_serial_number("k-s_n_w-123")
    'k-snw-123'
    >>> mangled_serial_number("-")
    Traceback (most recent call last):
        ...
    ValueError: '-'
    >>> mangled_serial_number("a-")
    Traceback (most recent call last):
        ...
    ValueError: 'a-'
    >>> mangled_serial_number("--")
    Traceback (most recent call last):
        ...
    ValueError: '-'
    >>> mangled_serial_number("-1-1")
    Traceback (most recent call last):
        ...
    ValueError: '-1-1'
    >>> mangled_serial_number("0-0")
    '0-0'
    >>> mangled_serial_number("0--------------0")
    '0-0'
    >>> mangled_serial_number("123-ix")
    '123-ix'
    >>> mangled_serial_number("123:ix")
    '123ix'
    >>> mangled_serial_number("myPmXYZ-123", strict=True)
    'mypmxyz-123'
    >>> mangled_serial_number("mYPmxyZ-12 3", strict=True)
    'mypmxyz-123'
    >>> mangled_serial_number("my PmxyZ-12 3", strict=True)
    'mypmxyz-123'
    >>> mangled_serial_number("m_y P_____________mxyZ-12 3", strict=True)
    'mypmxyz-123'
    >>> mangled_serial_number("myPmxyZ-------------------------12 3", strict=True)
    'mypmxyz-123'
    >>> mangled_serial_number("x-12345/1", strict=True)
    'x-123451'
    >>> mangled_serial_number("x-12345:1", strict=True)
    'x-123451'
    >>> import uuid
    >>> u_value = str(uuid.uuid4())
    >>> mangled_serial_number(u_value, strict=True) == u_value
    True
    """
    if value != 0 and not value:
        raise ValueError(repr(value))

    try:
        mangled = re.sub(REGEX_INVALID_CHARACTERS, "", value)
    except TypeError:
        mangled = re.sub(REGEX_INVALID_CHARACTERS, "", str(value))

    mangled = re.sub("\-+", "-", mangled)

    if not REGEX_VALID_SERIAL.match(mangled):
        raise ValueError(repr(mangled))

    if strict:
        return mangled.lower()

    return mangled


if __name__ == "__main__":
    import doctest
    import logging

    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    # logging.basicConfig(loglevel=logging.DEBUG)
    (FAILED, SUCCEEDED) = doctest.testmod()
    print("[doctest] SUCCEEDED/FAILED: {:d}/{:d}".format(SUCCEEDED, FAILED))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

.. seealso:

    * https://std.com/~dtd/sign_encrypt/sign_encrypt7.html
    * https://www.comparitech.com/blog/information-security/what-is-fernet/
    * https://pycryptodome.readthedocs.io/en/latest/src/examples.html#encrypt-data-with-rsa
    * https://cryptography.io/en/latest/hazmat/primitives/asymmetric/rsa/#signing
    * https://cryptography.io/en/latest/x509/tutorial/#determining-certificate-or-certificate-signing-request-key-type
    * https://cryptography.io/en/latest/x509/reference/#cryptography.x509.Certificate.tbs_certificate_bytes

"""
import struct
from collections import namedtuple
from typing import Union

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding, ec, rsa
from cryptography.fernet import Fernet
import cryptography
import cryptography.exceptions

MAGIC = 262866705
VERSION_TOW_FERNET = 1
VERSION_TOW_NACL = 2
VERSION_DEFAULT = VERSION_TOW_FERNET
SUPPORTED_VERSIONS = {1}

try:
    import nacl.secret
    import nacl.utils

    SUPPORTED_VERSIONS.add(VERSION_TOW_NACL)
except ImportError:
    pass

# magic, version, offset encrypted_session_key, offset symmetric_encrypted,
# offset inner signature, offset outer signature
BLOB_HEADER_FORMAT = f"!lBllll"

BLOB_HEADER_SIZE = struct.calcsize(BLOB_HEADER_FORMAT)

BlobPortions = namedtuple(
    "BlobPortions",
    (
        "encrypted_session_key",
        "symmetric_encrypted",
        "inner_signature_offset",
        "outer_signature",
        "header_data",
    ),
)

BlobHeaderPortions = namedtuple(
    "BlobHeaderPortions",
    (
        "magic",
        "version",
        "o_encrypted_session_key",
        "o_symmetric_encrypted",
        "o_inner_sig",
        "o_outer_sig",
    ),
)


class BlobUnsupportedVersion(ValueError):
    pass


class BlobInvalidSignatureInner(cryptography.exceptions.InvalidSignature):
    pass


class BlobInvalidSignatureOuter(cryptography.exceptions.InvalidSignature):
    pass


def mk_blob(
    encrypted_session_key: bytes,
    symmetric_encrypted: bytes,
    inner_signature_offset: int,
    outer_signature: bytes,
    version: int,
    **kwargs,
) -> bytes:
    encrypted_session_key_offset = BLOB_HEADER_SIZE
    symmetric_encrypted_offset = encrypted_session_key_offset + len(
        encrypted_session_key
    )
    outer_signature_offset = symmetric_encrypted_offset + len(
        symmetric_encrypted
    )

    blob = (
        struct.pack(
            BLOB_HEADER_FORMAT,
            MAGIC,
            version,
            encrypted_session_key_offset,
            symmetric_encrypted_offset,
            inner_signature_offset,
            outer_signature_offset,
        )
        + encrypted_session_key
        + symmetric_encrypted
        + outer_signature
    )

    return blob


def split_blob(blob: bytes) -> BlobPortions:
    blob_length = len(blob)
    header_data = BlobHeaderPortions(
        *struct.unpack(BLOB_HEADER_FORMAT, blob[:BLOB_HEADER_SIZE])
    )
    (
        r_magic,
        r_version,
        o_encrypted_session_key,
        o_symmetric_encrypted,
        o_inner_sig,
        o_outer_sig,
    ) = header_data

    assert r_magic == MAGIC
    assert r_version in SUPPORTED_VERSIONS
    assert BLOB_HEADER_SIZE == o_encrypted_session_key
    assert (
        o_encrypted_session_key
        < o_symmetric_encrypted
        < o_outer_sig
        < blob_length
    )

    return BlobPortions(
        blob[o_encrypted_session_key:o_symmetric_encrypted],
        blob[o_symmetric_encrypted:o_outer_sig],
        o_inner_sig,
        blob[o_outer_sig:],
        header_data,
    )


def mk_verifiable_blob(
    message: bytes,
    sender_key: rsa.RSAPrivateKey,
    recipient_key: rsa.RSAPublicKey,
    **kwargs,
) -> bytes:
    version = kwargs.get("version", VERSION_DEFAULT)
    assert version in SUPPORTED_VERSIONS

    if version == VERSION_TOW_FERNET:
        session_key = Fernet.generate_key()
        f = Fernet(session_key)
    elif version == VERSION_TOW_NACL:
        session_key = nacl.utils.random(nacl.secret.SecretBox.KEY_SIZE)
        box = nacl.secret.SecretBox(session_key)
    else:
        raise BlobUnsupportedVersion(version)

    inner_signature = sign_by(message + session_key, sender_key)
    inner_signature_offset = len(message)

    if version == VERSION_TOW_FERNET:
        symmetric_encrypted = f.encrypt(message + inner_signature)
    elif version == VERSION_TOW_NACL:
        symmetric_encrypted = box.encrypt(message + inner_signature)

    encrypted_session_key = encrypt_for(session_key, recipient_key)
    outer_signature = sign_by(
        encrypted_session_key + symmetric_encrypted, sender_key
    )

    return mk_blob(
        encrypted_session_key,
        symmetric_encrypted,
        inner_signature_offset,
        outer_signature,
        version=version,
    )


def decrypt_verifiable_blob(
    blob: bytes, sender_key: rsa.RSAPublicKey, recipient_key: rsa.RSAPrivateKey
):
    (
        encrypted_session_key,
        symmetric_encrypted,
        inner_signature_offset,
        outer_signature,
        header_data,
    ) = split_blob(blob)

    session_key = decrypt_for(encrypted_session_key, recipient_key)

    if header_data.version == VERSION_TOW_FERNET:
        f = Fernet(session_key)
        decrypted = f.decrypt(symmetric_encrypted)
    elif header_data.version == VERSION_TOW_NACL:
        box = nacl.secret.SecretBox(session_key)
        decrypted = box.decrypt(symmetric_encrypted)
    else:
        raise BlobUnsupportedVersion(header_data.version)

    message = decrypted[:inner_signature_offset]
    signature = decrypted[inner_signature_offset:]

    try:
        verify_signature_by(message + session_key, signature, sender_key)
    except cryptography.exceptions.InvalidSignature:
        raise BlobInvalidSignatureInner("Inner signature invalid!")

    try:
        verify_signature_by(
            encrypted_session_key + symmetric_encrypted,
            outer_signature,
            sender_key,
        )
    except cryptography.exceptions.InvalidSignature:
        raise BlobInvalidSignatureOuter("Outer signature invalid!")

    return message, session_key, signature


def decrypt_verifiable_blob_simple(
    blob: bytes, sender_key: rsa.RSAPublicKey, recipient_key: rsa.RSAPrivateKey
) -> bytes:
    message, session_key, signature = decrypt_verifiable_blob(
        blob, sender_key, recipient_key
    )

    return message


def encrypt_for(message: bytes, public_key: rsa.RSAPublicKey) -> bytes:
    """
    (RSA) Encryption of value for given key.

    Args:
        message (bytes): value
        public_key (rsa.RSAPublicKey): Recipient's public key

    Raises:
        ValueError: When no suitable public key was provided

    Returns:
        bytes: encrypted value
    """

    if isinstance(public_key, rsa.RSAPublicKey):
        return public_key.encrypt(
            message,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )

    raise ValueError(public_key)


def decrypt_for(message: bytes, private_key: rsa.RSAPrivateKey) -> bytes:
    """
    (RSA) decryption of value using given key.

    Args:
        message (bytes): encrypted value
        private_key (rsa.RSAPrivateKey): Recipient's private key

    Raises:
        ValueError: When no suitable private key was provided

    Returns:
        bytes: decrypted value
    """

    if isinstance(private_key, rsa.RSAPrivateKey):
        return private_key.decrypt(
            message,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )

    raise ValueError(private_key)


def sign_by(
    message: bytes,
    private_key: Union[rsa.RSAPrivateKey, ec.EllipticCurvePrivateKey],
) -> bytes:
    """
    Sign given value using private key.

    Args:
        message (bytes): value
        private_key (Union[rsa.RSAPrivateKey, ec.EllipticCurvePrivateKey]): private key

    Raises:
        ValueError: When no suitable private key was provided

    Returns:
        bytes: signature
    """

    if isinstance(private_key, rsa.RSAPrivateKey):
        return private_key.sign(
            data=message,
            padding=padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH,
            ),
            algorithm=hashes.SHA256(),
        )
    elif isinstance(private_key, ec.EllipticCurvePrivateKey):
        return private_key.sign(
            data=message,
            signature_algorithm=ec.ECDSA(hashes.SHA256()),
        )
    else:
        raise ValueError(private_key)


def verify_signature_by(
    message: bytes,
    signature: bytes,
    public_key: Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey],
) -> None:
    """
    Verify signature of given value using provided public key

    Args:
        message (bytes): value
        signature (bytes): signature
        public_key (Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey]): public key

    Raises:
        ValueError: When no suitable public key was provided

    """

    if isinstance(public_key, rsa.RSAPublicKey):
        return public_key.verify(
            signature=signature,
            data=message,
            padding=padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH,
            ),
            algorithm=hashes.SHA256(),
        )
    elif isinstance(public_key, ec.EllipticCurvePublicKey):
        return public_key.verify(
            signature=signature,
            data=message,
            signature_algorithm=ec.ECDSA(hashes.SHA256()),
        )
    else:
        raise ValueError(public_key)


if __name__ == "__main__":
    import doctest

    (FAILED, SUCCEEDED) = doctest.testmod()
    print(f"[doctest] SUCCEEDED/FAILED: {SUCCEEDED:d}/{FAILED:d}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
import logging
import traceback

import pendulum
import orjson


def log_traceback(message, exception, uselog=None):
    """
    Use *uselog* Logger to log a Traceback of exception *exception*.

    Args:
        message(str): message to be logged before trace log items
        exception(Exception): exception to be logged
        uselog(logging.Logger, optional): logger instance override

    """
    if uselog is None:
        uselog = logging.getLogger(__name__)
    e_type, e_value, e_traceback = sys.exc_info()

    uselog.warning(message)
    uselog.error(exception)

    for line in traceback.format_exception(e_type, e_value, e_traceback):
        for part in line.strip().split("\n"):
            if part != "":
                uselog.warning(part)


def dump_json(data, uselog=None, level=logging.INFO):
    """
    Log ``data`` in JSON format

    Args:
        data: data
        uselog(logging.Logger, optional): logger instance override
        level: logging level

    """
    if uselog is None:
        uselog = logging.getLogger(__name__)

    for line in orjson.dumps(
        data,
        option=orjson.OPT_INDENT_2 | orjson.OPT_APPEND_NEWLINE,
    ).split(b"\n"):
        uselog.log(level, line.decode("utf-8"))

def dump_json_to_stdout(data):
    """
    Dump ``data`` in JSON format to STDOUT

    Args:
        data: data

    """
    for line in orjson.dumps(
        data,
        option=orjson.OPT_INDENT_2 | orjson.OPT_APPEND_NEWLINE,
    ).split(b"\n"):
        print(line.decode("utf-8"))


def load_json(path):
    """
    Load JSON encoded file and return its contents.

    Args:
        path(str): path

    Returns:
        object: parsed content

    """
    with open(path, "rb") as src:
        content = orjson.loads(src.read())

    return content


def persist_json(data, path, indent=None, sort_keys=False):
    """
    Persist ``data`` in JSON format to ``path``.

    Args:
        data: data
        path: file path
        indent (int, optional): indent
        sort_keys (bool, optional): do sort keys

    """
    option = 0

    if indent:
        option |= orjson.OPT_INDENT_2

    if sort_keys:
        option |= orjson.OPT_SORT_KEYS

    with open(path, "wb") as tgt:
        tgt.write(orjson.dumps(data, option=option))


def vcrap(env_name=None):
    data = dict()

    try:
        data = orjson.loads(os.environ["VCAP_APPLICATION"])
    except Exception:
        pass

    space_name = data.get("space_name", "")

    if "prod" in space_name:
        env_name = "prod"
    elif "qa" in space_name:
        env_name = "qa"
    elif "dev" in space_name:
        env_name = "dev"

    if env_name:
        data["env_name"] = env_name

    return data


def period_object(period):
    """
    Convert *period* to  a ``pendulum.Period`` object if needed and possible.

    Args:
        period: input data

    Returns:
        period (pendulum.Period): period object
    """
    if isinstance(period, dict):
        try:
            try:
                return pendulum.Period(
                    pendulum.parse(period["begin"]),
                    pendulum.parse(period["end"]),
                )
            except KeyError:
                return pendulum.Period(
                    pendulum.parse(period["start"]),  # allowing also 'start'
                    pendulum.parse(period["end"]),
                )
        except KeyError:
            raise ValueError(
                "Cannot make a period object out of {!r}".format(period)
            )
    elif isinstance(period, pendulum.Period):
        return period

    raise ValueError("Cannot make a period object out of {!r}".format(period))


def period_json(period):
    """
    Convert a ``pendulum.Period`` object to a JSON serialisable dict.

    Args:
        period (pendulum.Period): period object

    Returns:
        dict: period specification
    """
    if isinstance(period, dict):
        period = period_object(period)

    return {
        "total_seconds": period.total_seconds(),
        "begin": period.start.to_iso8601_string(),
        "end": period.end.to_iso8601_string(),
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
try:
    try:
        from importlib.metadata import version
    except ImportError:
        from importlib_metadata import version
    __version__ = version(__package__)
except Exception:
    __version__ = "1.0.999"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Union

import pendulum
import orjson
from cryptography.x509.base import Certificate
from cryptography.hazmat.primitives.asymmetric.types import PRIVATE_KEY_TYPES

from coshiota.cryptography_lowlevel import VERSION_DEFAULT
from coshiota.cryptography_lowlevel import mk_verifiable_blob
from coshiota.cryptography_lowlevel import decrypt_verifiable_blob_simple


class TrustyPayloadMessage:
    def __init__(
        self,
        payload=None,
        sender_private: Union[PRIVATE_KEY_TYPES, None] = None,
        sender_public: Union[Certificate, None] = None,
        recipient_private: Union[PRIVATE_KEY_TYPES, None] = None,
        recipient_public: Union[Certificate, None] = None,
    ):
        if not payload:
            payload = dict()
        self.dt = pendulum.now()
        self.payload = payload
        self.sender_private = sender_private
        self.sender_public = sender_public
        self.recipient_private = recipient_private
        self.recipient_public = recipient_public

    def __str__(self):
        kind = "VIRGIN"
        a = ""
        b = ""

        if self.sender_private:
            kind = "TO:"

            try:
                a = f"{self.sender_public.subject} "
            except Exception:
                pass

            try:
                b = f" {self.recipient_public.subject}"
            except Exception:
                pass
        elif self.recipient_private:
            kind = "FROM:"

            try:
                a = f"{self.recipient_public.subject} "
            except Exception:
                pass

            try:
                b = f" {self.sender_public.subject}"
            except Exception:
                pass

        return f"<{self.__class__.__name__} {a}{kind}{b} {self.payload}>"

    def get_bytes(
        self,
        **kwargs,
    ) -> bytes:
        assert self.sender_private is not None
        assert self.recipient_public is not None

        version = kwargs.get("version", VERSION_DEFAULT)
        self.dt = pendulum.now()
        self.payload["dt"] = self.dt.timestamp()
        message = orjson.dumps(self.payload)

        return mk_verifiable_blob(
            message,
            self.sender_private,
            self.recipient_public.public_key(),
            version=version,
        )

    def parse_bytes(self, blob: bytes):
        assert self.sender_public is not None
        assert self.recipient_private is not None

        message = decrypt_verifiable_blob_simple(
            blob,
            self.sender_public.public_key(),
            self.recipient_private,
        )
        self.payload = orjson.loads(message)
        self.dt = pendulum.from_timestamp(self.payload["dt"])

    @classmethod
    def create(
        cls,
        payload: dict,
        sender_private: PRIVATE_KEY_TYPES,
        recipient_public: Certificate,
        sender_public: Union[Certificate, None] = None,
    ):
        return TrustyPayloadMessage(
            payload,
            sender_private=sender_private,
            sender_public=sender_public,
            recipient_public=recipient_public,
        )

    @classmethod
    def load(
        cls,
        blob: bytes,
        sender_public: Certificate,
        recipient_private: PRIVATE_KEY_TYPES,
        recipient_public: Union[Certificate, None] = None,
    ):
        obj = TrustyPayloadMessage(
            sender_public=sender_public,
            recipient_private=recipient_private,
            recipient_public=recipient_public,
        )
        obj.parse_bytes(blob)

        return obj



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import shlex


def parsed_cmdline():
    data = dict()

    with open("/proc/cmdline", "rt") as src:
        content = src.read().strip()

        for portion in shlex.split(content):
            key, value = portion, True
            if "=" in portion:
                key, value = portion.split("=", 1)
            data[key] = value

    return data
